"""
This module contains the assertions and error definitions.

Please see the specification for the errors.
- https://github.com/schul-cloud/schul_cloud_search_tests/blob/master/README.rst#specification
"""
from schul_cloud_resources_server_tests.tests.assertions import assertIsError
from schul_cloud_resources_server_tests.errors import errors
from schul_cloud_resources_api_v1.schema import get_schemas
from copy import deepcopy


ERROR_CLIENT_REQUEST = 400 # https://httpstatuses.com/400
Q = "Q" # The query string
ERROR_SERVER_RESPONSE = 409 # https://httpstatuses.com/409

ERROR = get_schemas()["error"].get_valid_examples()[0]


def get_error(error_code):
    """Return an error with the specified code."""
    error = deepcopy(ERROR)
    error["errors"][0]["status"] = str(error_code)
    error["errors"][0]["title"] = errors[int(error_code)]
    return error


def assertServerReplyIsWrong(response):
    """Make sure the response is because the server responded in an
    unspecified manner.
    """
    assert response.status_code == ERROR_SERVER_RESPONSE
    assertIsError(response, ERROR_SERVER_RESPONSE)


def assertClientRequestIsInvalid(response):
    """Make sure the response is an error in the client request.
    """
    assert response.status_code == ERROR_CLIENT_REQUEST
    assertIsError(response, ERROR_CLIENT_REQUEST)



from pytest import mark
from schul_cloud_search_tests.tests.assertions import (
    assertIsError, ERROR_SERVER_RESPONSE, ERROR)
import copy
from pprint import pprint

INVALID_ACCEPT_HEADERS = [
        "application/vnd.api+json; v=1",
        "application/json",
        "application/json,application/vnd.api+json; v=1,",
        "image/jpeg",
        ""
    ]


@mark.parametrize("invalid_accept_header", INVALID_ACCEPT_HEADERS)
def test_detect_invalid_accept_headers(
        search_engine, invalid_accept_header):
    """Test that the server responds to invalid media type queries.
    
    This test currently only tests that the server MUST send a 406 error
    if the media type application/vnd.api+json is not accepted by the client.
    Usually, an other media type is also allowed to change the result,
    e.g. application/xml or application/json.
    This test assumes that the server speaks only application/vnd.api+json.
    
    See
    - http://jsonapi.org/format/#content-negotiation-servers
    
    The server responds with 200 but 406 is expected.
    """
    result = search_engine.request(headers={"Accept":invalid_accept_header})
    assertIsError(result, ERROR_SERVER_RESPONSE)


@mark.parametrize("invalid_accept_header", INVALID_ACCEPT_HEADERS)
def test_server_responds_with_406_and_is_passed_through(
        search_engine, invalid_accept_header):
    """When the client requests invalid headers, a 406 is passed through.
    """
    error406 = copy.deepcopy(ERROR)
    error406["errors"][0]["status"] = "406"
    error406["errors"][0]["title"] = "Not Acceptable"
    response = search_engine.host(error406, status_code=406).request(
                   headers={"Accept":invalid_accept_header})
    result = response.json()
    assert result == error406


@mark.parametrize("valid_accept_header", [
        "application/vnd.api+json; v=1,application/vnd.api+json,application/json",
        "application/vnd.api+json,application/json",
        "application/vnd.api+json",
        "application/*",
        "*/*",
    ])
def test_allow_list_of_accept_headers(search_engine, valid_accept_header):
    """If a list of accept headers is passed to the server, it can accept.
    
    The client may pass a list of media type parameters to the server.
    The server finds out that a valid parameter is included.
    See
    - http://jsonapi.org/format/#content-negotiation-servers
    """
    response = search_engine.request(headers={"Accept":valid_accept_header})
    result = response.json()
    assert result == search_engine.last_response


@mark.parametrize("headers", [
        {"Xsss": "23423"},
        {"Accept": "application/json"},
        {"Authorization": "basic basic===", "X-Asd": "22"},
    ])
def test_search_engine_passes_headers_through(search_engine, headers):
    """Test that certain headers pass through the proxy.
    
    """
    search_engine.request(headers=headers)
    for key, value in headers.items():
        assert search_engine.last_request_headers[key] == value


@mark.parametrize("headers", [
        {"Content-Type": "application/vnd.api+json; charset=UTF-8"},
        {"Content-Type": "application/vnd.api+json; Y2736487=sdfsdf"},
        {"Content-Type": "application/vnd.api+json; charset=UTF-8,jsdhfgaskfgdskgfkjas"},
        {"Content-Type": "application/vnd.api+json; charset=UTF-8sjhfkjafsd"},
        {"Test-Header": "123"}
    ])
def test_response_content_type_headers_are_accepted(search_engine, headers):
    """Make sure that different returned content types are accepted by the proxy.
    
    See Issue 11
        https://github.com/schul-cloud/schul_cloud_search_tests/issues/11
    
    > Clients MUST ignore any parameters for the application/vnd.api+json media type received in the Content-Type header of response documents.
    - http://jsonapi.org/format/#content-negotiation-clients
    """
    response = search_engine.host(headers=headers).request()
    result = response.json()
    pprint(result)
    assert result == search_engine.last_response
    for key, value in headers.items():
        assert response.headers[key] == value





"""
This file contains the test configuration like fixtures.
"""
from pytest import fixture
from schul_cloud_resources_server_tests.tests.fixtures import ParallelBottleServer
from bottle import Bottle, request, response
import sys
import os
import requests
from urllib.parse import urlencode
from copy import deepcopy
from schul_cloud_search_tests.tests.assertions import (
    Q, assertClientRequestIsInvalid, get_error, ERROR_CLIENT_REQUEST)
import json
from schul_cloud_resources_api_v1.schema import get_schemas
from collections import namedtuple
import pytest


HERE = os.path.dirname(__file__)
MODULE_ROOT = os.path.join(HERE, "..", "..")
try:
    import schul_cloud_search_tests
except ImportError:
    sys.path.append(MODULE_ROOT)
from schul_cloud_search_tests.proxy import get_app


DEFAULT_PARAMETERS = {"Q":"test"}
DEFAULT_HEADERS = {}
DEFAULT_STATUS_CODE = 200
JSONAPI = {
    "version": "1.0",
    "meta" : {
      "name": "Example Server",
      "source": "https://github.com/schul-cloud/resources-api-v1",
      "description": "This is a test server for the search API."
    }
}


class Requester(object):
    """Shortcut to request a hosted resource."""
    
    def __init__(self, url):
        """Create a requester."""
        self._url = url
    
    def request(self, headers={}):
        """Request the resource."""
        return requests.get(self._url, headers=headers)


def ending_with_slash(url):
    """Return the url with a slash in the end."""
    if not url.endswith("/"):
        url = url + "/"
    return url


def params_to_key(params):
    """Return a tuple key for the parameters."""
    for key, value in params.items():
        assert isinstance(key, str)
        assert isinstance(value, str)
    key = list(params.items())
    key.sort()
    return tuple(key)


class SearchEngine(object):
    """The search engine adapter.
    
    This includes:
    - the running search tests server
    - a bottle server returning the registeres responses
    """
    
    _default_response = {
      "jsonapi": JSONAPI,
      "links": {
        "self": {
          "href": None,
          "meta": {
            "count": 0,
            "offset": 0,
            "limit": 0
          }
        },
        "first": None,
        "last": None,
        "prev": None,
        "next": None
      },
      "data": []
    }

    def __init__(self):
        """Create a search engine object."""
        self._queries = {}
        self._app = Bottle()
        self._app.get("/", callback=self._serve_request)
        self._reset_servers()
    
    def start(self):
        """Start the search engine tests and the search engine mock."""
        assert not self.is_started()
        self._server = ParallelBottleServer(self._app)
        self._search_app = get_app(target_url=self.search_engine_url)
        self._search_server = ParallelBottleServer(self._search_app)
        
    @property
    def search_engine_url(self):
        """Return the URL of the mocked search engine."""
        assert self.is_started()
        return ending_with_slash(self._server.url)

    @property
    def proxy_url(self):
        """Return the url of the search engine tests proxy."""
        assert self.is_started()
        return ending_with_slash(self._search_server.url)
        
    def stop(self):
        """Stop the search engine tests and the mock server."""
        assert self.is_started()
        self._server.shutdown()
        self._search_server.shutdown()
        self._reset_servers()

    def _reset_servers(self):
        """Create the starting condition"""
        self._server = None
        self._search_server = None
        self._search_app = None

    def is_started(self):
        """Return whether the server is started."""
        return self._server is not None
        
    def host(self,
             response=None,
             params=DEFAULT_PARAMETERS,
             headers=DEFAULT_HEADERS,
             status_code=DEFAULT_STATUS_CODE):
        """Host the response given by the query.
        
        Default headers are 
        - "Content-Type": "application/vnd.api+json"
        """
        if response is None:
            response = self.get_default_response(params)
        key = params_to_key(params)
        print("host:", key, file=sys.stderr)
        assert key not in self._queries
        self._queries[key] = (response, headers, status_code)
        return self._new_requester(params)
    
    def clear(self):
        """Remove all hosted responses."""
        self._queries = {}
        self._last_response = None
        self._last_request_headers = None

    def _serve_request(self):
        """Serve a request to the search engine bottle server."""
        default = object()
        key = params_to_key(request.query)
        print("request:", key, file=sys.stderr)
        self._last_request_headers = dict(request.headers)
        result = self._queries.get(key, default)
        if result is default:
            json_response = self.get_default_response(request.query_string)
            headers = DEFAULT_HEADERS
            status_code = DEFAULT_STATUS_CODE
            print("Return default result.", file=sys.stderr)
        else:
            json_response, headers, status_code = result
        self._last_response = json_response
        default_headers = {"Content-Type": "application/vnd.api+json"}
        default_headers.update(headers)
        for header, value in default_headers.items():
            response.set_header(header, value)
        response.status = status_code
        if isinstance(json_response, dict):
            return json.dumps(json_response).encode("UTF-8")
        return json_response

    def request(self,
                params=DEFAULT_PARAMETERS,
                headers=DEFAULT_HEADERS):
        """Request a search with parameters."""
        return self._new_requester(params).request(headers)
        
    def _new_requester(self, params):
        """Return a new Requester object from the parameters."""
        request_url = self.proxy_url + "?" + urlencode(params)
        return Requester(request_url)
        
    def get_default_response(self, query):
        """Return the default response to a query.
        
        The query can be given as string or as dict.
        """
        if not isinstance(query, str):
            query = urlencode(query)
        result = deepcopy(self._default_response)
        url = self.search_engine_url + "?" + query
        result["links"]["self"]["href"] = url
        return result
        
    @property
    def last_response(self):
        """Return the last reponse of the search engine to a request.
        
        If there was no request, None is returned.
        """
        return self._last_response

    @property
    def last_request_headers(self):
        """Return the last request headers of the search engine in a request.
        
        If there was no request, None is returned.
        """
        return self._last_request_headers


@fixture(scope="session")
def search_engine_session():
    """Return the server to store resources."""
    server = SearchEngine()
    server.start()
    yield server
    server.stop()


@fixture
def search_engine(search_engine_session):
    """Return a fresh server object with no resources."""
    search_engine_session.clear()
    yield search_engine_session
    search_engine_session.clear()


@fixture
def code_url(search_engine):
    """Return the url of the code enpoint."""
    return search_engine.proxy_url + "code"


VALID_RESOURCE = get_schemas()["resource"].get_valid_examples()[0]
@fixture
def valid_resource():
    """Return a valid resource."""
    return {"type": "resource", "attributes": deepcopy(VALID_RESOURCE), "id": "1"}


@fixture(params=[11, 20])
def limit(request):
    """Return the limit of one search request."""
    return request.param


@fixture(params=[0, 8, 100])
def link_resources(request, valid_resource, limit):
    """Return a list of resources wich are returned by a search."""
    result = [[]]
    count = 0
    count_in_last_request = 0
    while count < request.param:
        if count_in_last_request >= limit:
            result.append([])
            count_in_last_request = 0
        result[-1].append(valid_resource)
        count += 1
        count_in_last_request += 1
    return result


@fixture
def link_parameters(link_resources, limit):
    """Return the parameters used to search."""
    result = [{Q:"tralala"}]
    count = len(link_resources[0])
    for res in link_resources[1:]:
        result.append({Q:"tralala", "page[limit]":str(limit), "page[offset]":str(count)})
        count += len(res)
    return result


@fixture
def link_urls(search_engine, link_parameters):
    """Return a list of links for the paramaters"""
    return [search_engine.search_engine_url + "?" + urlencode(params)
            for params in link_parameters]


@fixture
def link_responses(link_urls, link_parameters, link_resources, limit):
    """Return a response with filled links.
    
    The response is not hosted by the search engine.
    """
    result = []
    offset = 0
    for url, parameters, resources, next_url, previous_url in zip(
            link_urls, link_parameters, link_resources,
            link_urls[1:] + [None], [None] + link_urls[:-1]):
        response = {
          "jsonapi": JSONAPI,
          "links": {
            "self": {
              "href": url,
              "meta": {
                "count": len(resources),
                "offset": offset,
                "limit": limit,
              }
            },
            "first": (link_urls[0] if resources else None),
            "last": (link_urls[-1] if resources else None),
            "prev": previous_url,
            "next": next_url,
          },
          "data": resources,
        }
        result.append(response)
        offset += len(resources)
    return result

LinkedRequest = namedtuple("LinkedRequest", [
    "resources", "parameters", "url", "response", "request"])

@fixture
def link_request(search_engine, link_parameters, link_responses):
    """Return a list of requesters for the responses."""
    return [search_engine.host(response, parameters).request
            for response, parameters in zip(link_responses, link_parameters)]


@fixture
def linked_search(link_responses, link_urls, link_parameters, link_resources, 
          link_request):
    """Return a list of linked resource requests.
    
    Each element has the following attributes:
    - response
    - url
    - paramaters
    - resources
    """
    return list(map(lambda t: LinkedRequest(*t), zip(link_resources, link_parameters, link_urls, link_responses, link_request)))


@fixture
def first_search(linked_search):
    """The first of the linked searches."""
    return linked_search[0]


@fixture
def second_search(linked_search):
    """The second of the linked searches."""
    if len(linked_search) < 2:
        pytest.skip("Not enough searches with a limit.")
    return linked_search[1]


@fixture
def third_search(linked_search):
    """The third of the linked searches."""
    if len(linked_search) < 3:
        pytest.skip("Not enough searches with a limit.")
    return linked_search[2]


@fixture
def last_search(linked_search):
    """The last of the linked searches."""
    return linked_search[-1]


@fixture
def high_offset_search(last_search, first_search):
    """Return a search with an offset which is too high."""
    if last_search == first_search:
        pytest.skip("need mutiple searches")
    last_search.response["data"] = []
    last_search.response["links"]["self"]["meta"]["count"] = 0
    last_search.response["links"]["next"] = None
    last_search.response["links"]["prev"] = None
    last_search.response["links"]["last"] = None
    last_search.response["links"]["first"] = None
    return last_search


@fixture
def not_last_search(first_search, last_search):
    """Return a search which is definitely not the last search."""
    if first_search == last_search:
        pytest.skip("I need a request that is not the last.")
    return first_search


@fixture(params=[
        get_error(ERROR_CLIENT_REQUEST),
        None
    ])
def response400(request):
    """Return the response which yeilds a 400 by the proxy."""
    return request.param


@fixture
def assertIs400(response400):
    """Return an assertion function which tests that the result is correct."""
    if response400 is None:
        return assertClientRequestIsInvalid
    else:
        def test_error_passes_through(response):
            """Make sure the error passes through."""
            response = response.json()
            print("response:", response)
            print("response400:", response400)
            assert response == response400
        return test_error_passes_through


@fixture
def request400(search_engine, response400):
    """Return a function that requests parameters."""
    return lambda params: search_engine.host(response=response400, params=params, status_code=(200 if response400 is None else 400)).request()



"""
Each test function here covers a part of the specification.
They are run when the server returns a request.
If they fail, the client gets feedback.

About the test functions:
Please provide
- A short description of what is tested
- Links to the specification, where this is specified.

"""
from pytest import mark
from schul_cloud_resources_api_v1.schema import get_schemas
from schul_cloud_resources_server_tests.tests.assertions import assertIsError


def test_result_is_json(result):
    """Make sure the result is a json object.
    
    See
    - http://jsonapi.org/"""
    try:
        json = result.json()
    except:
        print("Result:", result.content)
        assert False, "The result must be a valid json object, not {}".format(result.content)
    else:
        assert isinstance(json, dict), "The result must be a json object, not {}".format(json)


def test_response_schema(search_response):
    """If the request was successful, it must match the search-response schema.
    
    Schema:
    - https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/search-response
    
    The response is successful of a status code 200 is returned.
    """
    schema = get_schemas()["search-response"]
    schema.validate(search_response)
    

def test_400_and_500_status_codes_have_the_jsonapi_design(error, result):
    """If the status code is 4XX, it contains a list of errors.

    Schema:
    - https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/error
    """
    assertIsError(error, result.status_code)
    

def test_the_content_type_is_from_the_jsonapi(result):
    """Make sure the returned content type is specified and expected.
    
    See here:
    - http://jsonapi.org/format/#content-negotiation-clients
    """
    content_type = result.headers.get("Content-Type", "")
    content_type_without_parameters = content_type.split(";", 1)[0]
    assert content_type_without_parameters == "application/vnd.api+json", "The content type of the server reply must be application/vnd.api+json."


def test_406_error_is_expected_in_case_of_invalid_accept_headers(
        search, result, json):
    """If the client does not acccept application/vnd.api+json, 
    Error 406 should be returned.

    See
    - http://jsonapi.org/format/#content-negotiation-servers
    """
    accept = search.headers.get("Accept", None)
    if accept is None: return
    accepted_content_types = accept.split(",")
    print("Accepted content types:", accepted_content_types)
    possible_content_types = ["*/*", "application/*", "application/vnd.api+json"]
    content_type_is_accepted = any(
        possible_content_type in accepted_content_types
        for possible_content_type in possible_content_types)
    must_be_406 = not content_type_is_accepted
    is_406 = result.status_code == 406
    assert is_406 == must_be_406, "Error 406 is returned only if the content type is not accepable."



"""
Each test function here covers a part of the specification.
They are run when the server returns a request.
If they fail, the client gets feedback.

About the test functions:
Please provide
- A short description of what is tested
- Links to the specification, where this is specified.

These test functions all relate to the "links" attribute of the search response.
- https://github.com/schul-cloud/resources-api-v1/tree/master/schemas/search-response#search-reponse-1

"""
from pytest import mark
from urllib.parse import parse_qs, urlparse


def get_offset(link):
    """Return the offset from a link."""
    query = parse_qs(urlparse(get_href(link)).query)
    return int(query.get("page[offset]", [0])[0])


def get_href(link):
    """Return the href of the link."""
    if link is None:
        return None
    if isinstance(link, dict):
        return link.get("href")
    assert isinstance(link, str), "a link must be either a string or an object."
    return link


def test_count_is_less_or_equal_to_the_limit(self_link):
    """The limit is the maximum number of objects returned.
    """
    assert self_link["meta"]["count"] <= self_link["meta"]["limit"]


def test_limit_may_be_reduced(self_link, limit):
    """If the limit is higher than the maximum limit the server has internally, the server sets the limit to the maximum available limit.
    """
    assert limit is None or self_link["meta"]["limit"] <= limit


def test_count(self_link, search_response, data):
    """count is the actual number of objects retrieved.
    
    The count must be equal to the number of objects in the list.
    """
    assert len(data) == self_link["meta"]["count"]


def test_offset(self_link, offset):
    """offset is the start index in the list of objects.
    
    The offset must be the same as requested.
    When a query is done without offset, the requested offset is zero.
    """
    assert self_link["meta"]["offset"] == offset


def test_the_end_is_reached(self_link, links):
    """If the end of the resource list is reached, count may be less than limit.
    
    This implies that 
    - there is no next link and 
    - if there are resources, the last link is self 
    - if there are no resources in this request
      - the last link is pointing to a lesser offset or
      - the last link is null
    """
    if self_link["meta"]["count"] and \
            self_link["meta"]["count"] < self_link["meta"]["limit"]:
        assert links["next"] is None
        assert self_link["href"] == get_href(links["last"])
    elif self_link["href"] == get_href(links["last"]):
        assert links["next"] is None


def test_no_resources_given(search_response, links, offset):
    """If there are no resources, first, last, prev and next MUST be null.
    """
    if len(search_response["data"]) == 0:
        if offset == 0:
            assert links["last"] is None
            assert links["first"] is None
        assert links["prev"] is None
        assert links["next"] is None


def test_last_implies_next(links, offset):
    """If links["last"] is given, links["next"] must be given.

    This only works if this is not the last request, becuase
    the last request has no next request.
    """
    if links["last"] and get_offset(links["last"]) > offset:
        print("last offset:", get_offset(links["last"]))
        assert links["next"] is not None, "The next link must be given if self is before the last link."


def test_first_implies_prev(links, offset):
    """If first is given, prev must be given.
    
    This is only valid, if we did not request an offset which is out of range.
    """
    if links["first"] and links["last"] \
            and get_offset(links["last"]) > offset > 0:
        assert links["prev"] is not None, "The prev link must be given if there is a first link and the request is not out of range."


def test_next_does_not_skip_objects(links, self_link):
    """The next link MUST not skip objects.
    
    This can only be inferred by the limit and offset.
    """
    if links["next"]:
        next_offset = get_offset(links["next"])
        self_offset = self_link["meta"]["offset"]
        limit = self_link["meta"]["limit"]
        assert next_offset == self_offset + limit, "The offset of the next link must be increased by the limit."
        

def test_prev_does_not_skip_objects(links, self_link):
    """The prev link MUST not skip objects.
    
    This can only be inferred by the limit and offset.
    """
    if links["prev"]:
        prev_offset = get_offset(links["prev"])
        self_offset = self_link["meta"]["offset"]
        limit = self_link["meta"]["limit"]
        assert prev_offset == self_offset - limit, "The offset of the prev link must be decreased by the limit."


@mark.parametrize("link_name", ["prev", "next", "last", "first", "self"])
def test_links_are_absolute(link_name, links):
    """All links are absolute links using the host header field.
    """
    href = get_href(links[link_name])
    if href is not None:
        link_scheme = urlparse(href).scheme
        print("link_scheme:", repr(link_scheme), href)
        assert link_scheme in ["http", "https"], "there must be a scheme like http:// in the url for {}, not {} in ".format(link_name, repr(link_scheme))


def test_request_too_far(links, offset, self_link, data):
    """Requesting beyond the last link yields no objects.
    
    A request beyond the last possible request yields no objects.
    """
    limit = self_link["meta"]["limit"]
    if links["last"] and get_offset(links["last"]) + limit <= offset:
        assert len(data) == 0, "No data should be given but {} objects were found.".format(len(data))



"""
Test the headers of search requests.
"""


from schul_cloud_search_tests.tests.test_headers import (
    INVALID_ACCEPT_HEADERS)
from pytest import mark


@mark.parametrize("invalid_accept_header", INVALID_ACCEPT_HEADERS)
def test_detect_invalid_accept_headers(validateRequest, invalid_accept_header):
    validateRequest(headers={"Accept":invalid_accept_header})





"""
This file contains the fixtures used in the tests in the same folder.

"""
from pytest import fixture
from urllib.parse import parse_qs, urlencode, urlparse
import requests
import sys
import os
from schul_cloud_resources_server_tests.tests.fixtures import (
    ParallelBottleServer)
import time
from pprint import pprint

HERE = os.path.dirname(__file__)
MODULE_ROOT = os.path.join(HERE, "..", "..")
try:
    import schul_cloud_search_tests
except ImportError:
    sys.path.append(MODULE_ROOT)
from schul_cloud_search_tests.proxy import get_app


DEFAULT_QUERIES = [{"Q":"einstein"}, {"Q":"test"}]


def pytest_addoption(parser):
    """Add options to pytest.

    This adds the options for
    - url to store the value
    - query to add the token to a list
    - max-depth to store the value
    """
    parser.addoption("--url", action="store",
        default="http://localhost:8080/v1/search",
        help="url: the url of the server api to connect to")
    parser.addoption("--max-depth", action="store", default=20,
        help="max-depth: This limits the number of requests"
             "which are executed to get all search results.")
    parser.addoption("--query", action="append", default=[],
        help="query: list of queries to ask the search engine")


def pytest_generate_tests(metafunc):
    """Generate parameters.

    - query
        These are some query parameters passed over via the command line
        
        If not enough parameters are specified, a set of own paramaters is taken.
        They can be specified by the command line parameter --query.
    """
    if "query"  in metafunc.fixturenames:
        queries = list(map(
            lambda qs:
                {param:values[1] for param, values in parse_qs(qs).items()},
            metafunc.config.option.query))
        queries += DEFAULT_QUERIES[len(queries):]
        metafunc.parametrize("query", queries)


@fixture(scope="session")
def search_url(request):
    """This is the url of the search engine to connect to.
    
    The url is passed as the parameter --url.
    """
    return request.config.getoption("--url")


@fixture(scope="session")
def secret():
    return time.time()


@fixture
def validateRequest(search_url, search_tests_url, secret):
    """Return a function to request results from the search engine.
    
    The results are validated.
    If it happens that the search engine does not provide a valid response, 
    the response is printed and an AssertionError is raised.
    """
    def validateRequest(query=DEFAULT_QUERIES[0], headers={}):
        if isinstance(query, dict):
            url = search_tests_url + "?" + urlencode(query)
        elif isinstance(query, str):
            if query.startswith(search_tests_url):
                url = search_tests_url + "?" + urlparse(query).query
            else:
                assert query.startswith(search_url)
                url = search_tests_url + query[len(search_url):]
        else:
            raise TypeError("query must be either dict or str.")
        h = {"Content-Type": "application/vnd.api+json"}
        h.update(headers)
        result = requests.get(url, headers=h)
        try:
            result_json = result.json()
        except:
            print("Requesting", query, "->", url)
            raise
        if result_json["jsonapi"]["meta"].get("secret") == secret:
            for error in result_json["errors"]:
                 meta = error.get("meta")
                 if meta:
                     print(meta.get("traceback"))
                     print(error.get("detail"))
                     print(meta.get("github-url"))
                 else:
                    print("Error:", error.get("detail"))
                 print()
            assert False, "search engine tests failed."
        return result
    return validateRequest


@fixture(scope="session")
def search_tests_url(search_url, secret):
    """The url of the search engine tests."""
    search_app = get_app(target_url=search_url, secret=secret)
    search_server = ParallelBottleServer(search_app)
    yield search_server.url
    search_server.shutdown()


@fixture(scope="session")
def max_depth(request):
    """Return a number of requests to perform at maximum.
    
    This limits the number of requests which are executed
    to get all search results.
    The can be specified by the command line parameter --max-depth.
    """
    return int(request.config.getoption("--max-depth"))



"""
This file provides the tests which should be run against a running search engine.

You can view examples of such tests in the Issue 6.
    https://github.com/schul-cloud/schul_cloud_search_tests/issues/6

The idea is to use these tests to verify the behavior of the search api.
    https://github.com/schul-cloud/resources-api-v1#search-api
"""
from schul_cloud_search_tests.tests.test_request import (
    MALFORMED_PARAMETERS)
from pytest import mark


def test_q_is_missing(validateRequest):
    """Request a search with the q parameter missing."""
    validateRequest({})


@mark.parametrize("parameter", [parameter for 
                                parameter, parameter_is_valid in
                                MALFORMED_PARAMETERS if not parameter_is_valid])
def test_malformed_parameters(validateRequest, parameter):
    """Test that the list of malformed parameters results in a 400 answer."""
    validateRequest({parameter:"0"})


@mark.parametrize("param", ["page[offset]", "page[limit]"])
@mark.parametrize("value", ["", "asd", "-1"])
def test_invalid_paramater_values(validateRequest, param, value):
    """Test that the search engine recognizes invalid parameters.
    
    The answer 400 is expected.
    """
    validateRequest({param:"0"})














"""
These tests are run by the proxy.
They test whether requests of the client and responses of the server are correct.

The tests in schul_cloud_search_tests.tests test that these tests work.

"""
import pytest
import os
import threading
import io
import traceback
import inspect
from bottle import request

HERE = os.path.dirname(__file__)
_local = threading.local()
SOURCE_BASE = "https://github.com/schul-cloud/schul_cloud_search_tests/tree/master/schul_cloud_search_tests"


def get_test_function(tb):
    """Return the documentation from the first test functions found."""
    while tb:
        for func in tb.tb_frame.f_globals.values():
            if inspect.isfunction(func) and \
                    func.__name__.startswith("test_") and \
                    func.__code__ == tb.tb_frame.f_code:
                return func
        tb = tb.tb_next


def run_request_tests(request_url):
    """Run the tests with the request."""
    return _run_tests_and_collect_errors(request_url, ["-m" "request", HERE])


def run_response_tests(request_url, response):
    """Run the tests with the request and response."""
    _local.response = response
    try:
        return _run_tests_and_collect_errors(request_url, ["-m" "not request", HERE])
    finally:
        del _local.response


def get_response():
    """Return the currently working response."""
    return _local.response


def get_request_url():
    """Return the url of the test server endpoint with the query."""
    return _local.request_url


def get_source_url(function, base=None):
    """Return the source url of the function where the source code resides."""
    if base is None:
        host = request.headers.get("host", None)
        if host is None:
            url = SOURCE_BASE
        else:
            url = "http://" + host + "/code"
    else:
        url = base
    base = os.path.abspath(os.path.dirname(HERE))
    path = os.path.abspath(function.__globals__.get("__file__", ""))
    if not path.startswith(base):
        return None
    return url + path[len(base):].replace("\\", "/") + "#L" + str(function.__code__.co_firstlineno)


def add_failing_test(ty, err, tb):
    """Add a failing test to the tests of this request."""
    file = io.StringIO()
    traceback.print_exception(ty, err, tb, file=file)
    tb_string = file.getvalue()
    test = get_test_function(tb)
    _local.failing_tests.append({
          "status": 500, 
          "title": "Internal Server Error",
          "detail": ty.__name__ + ": " + str(err),
          "meta": {
            "traceback" : tb_string,
            "documentation": (test.__doc__ if test else None),
            "test_function": (test.__name__ if test else None),
            "source": (get_source_url(test) if test else None),
            "source-line": (test.__code__.co_firstlineno if test else None),
            "error-class": ty.__module__ + "." + ty.__name__,
            "error": str(err),
            "github-url": (get_source_url(test, SOURCE_BASE) if test else None),
          }
        })


def _run_tests_and_collect_errors(request_url, args):
    """Run the tests and collect the errors in these tests."""
    _local.failing_tests = []
    _local.request_url = request_url
    try:
        pytest.main(args)
        return _local.failing_tests
    finally:
        del _local.failing_tests
        del _local.request_url




"""
Run tests for the requests and responses.

If you mark your test with
@pytest.mark.request
your test is executed when the request arrives.

"""

from pytest import fixture, hookimpl
import pytest
from schul_cloud_search_tests.search_tests import (
    get_response, add_failing_test, get_request_url
)
import pytest


try:
    pytest.skip()
except Exception as e:
    SKIP_ERROR = e.__class__


@fixture
def search():
    """The request to the server.
    
    The request from the client
    - http://bottlepy.org/docs/dev/api.html#bottle.BaseRequest
    """
    from bottle import request
    return request


@fixture
def result():
    """The response to the query.
    
    The response from the server.
    - http://docs.python-requests.org/en/master/api/#requests.Response
    """
    return get_response()


@fixture
def json(result):
    """Return the response as json."""
    try:
        return result.json()
    except:
        pytest.skip("The result is not a valid json object. {}"
                           .format(result.content))


@fixture
def search_response(json, result):
    """Return the search response."""
    if not result.ok:
        pytest.skip("This is not a search response.")
    return json


@fixture
def error(json, result):
    """Return the error response."""
    if result.ok:
        pytest.skip("This is not an error response.")
    return json


@fixture
def search_url():
    """The url of the server to request searches from."""
    return get_request_url()


@fixture
def params(search):
    """The query parameters.
    
    The search request parameters:
    - http://bottlepy.org/docs/dev/api.html#bottle.FormsDict
    
    You can use params.get("q") to get the "q" parameter.
    """
    return search.query


@fixture
def offset(params):
    """Return the requested offset."""
    return int(params.get("page[offset]", "0"))


@fixture
def limit(params):
    """Return the requested limit."""
    limit = params.get("page[limit]")
    return (None if limit is None else int(limit))


@fixture
def links(search_response):
    """Return the links of the search response."""
    return search_response["links"]


@fixture
def self_link(links):
    """Return the self link."""
    return links["self"]


@fixture
def data(search_response):
    """Return the data field of the search response."""
    return search_response["data"]

@hookimpl(hookwrapper=True)
def pytest_pyfunc_call(pyfuncitem):
    # From
    # - https://docs.pytest.org/en/latest/writing_plugins.html#conftest-py-plugins
    # Also see 
    # - https://docs.pytest.org/en/latest/_modules/_pytest/vendored_packages/pluggy.html#_CallOutcome
    # for outcome
    
    outcome = yield
    # outcome.excinfo may be None or a (cls, val, tb) tuple
    if outcome.excinfo is not None and outcome.excinfo[0] not in (SKIP_ERROR,):
        add_failing_test(*outcome.excinfo)
    res = outcome.get_result()



from bottle import get, run, response, request
import json
from schul_cloud_resources_api_v1.schema import get_schemas


@get("/<path:path>")
def get(path):
    """Return a search response."""
    response.set_header("Content-Type", "application/vnd.api+json")
    self_href = "http://" + request.headers.get("host", "localhost") + "?" + request.query_string
    example_resources = get_schemas()["resource"].get_valid_examples()
    resources = [{"type":"resource","attributes":data,"id":str(i)}
                 for i, data in enumerate(example_resources)]
    result = {
        "jsonapi": {
            "version": "1.0",
            "meta": {
                "name": "schul_cloud/schul_cloud_search_tests", 
                "source": 
                    "https://github.com/schul_cloud/schul_cloud_search_tests",
                "description":
                    "These are the tests for the search engines.",
            }
        },
        "links" : {
            "self": {
                 "href": self_href,
                 "meta": {
                    "limit": 10,
                    "offset": 0,
                    "count": len(resources)
                 }
            },
            "first": self_href,
            "last": self_href,
            "prev": None,
            "next": None,
        },
        "data": resources
    }
    return json.dumps(result)

if __name__ == "__main__":
    run(reload=True, port=8080)



"""
Run the tests agains a search engine.
"""

import os
import sys
import pytest

HERE = os.path.dirname(__file__)
API_TESTS = HERE
arguments = sys.argv[1:]
if len(arguments) > 0 and not arguments[0].startswith("-"):
    arguments.insert(0, "--url")

errcode = pytest.main([API_TESTS] + arguments)
sys.exit(errcode)



from bottle import (
    Bottle, request, response, redirect, static_file, SimpleTemplate
)
import sys
import os
import shutil
import json
import requests
try:
    from json import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError
from schul_cloud_resources_server_tests.tests.fixtures import StoppableWSGIRefServerAdapter
if "" in sys.path:
    sys.path.append(".")
from schul_cloud_search_tests.search_tests import run_request_tests, run_response_tests

ENDPOINT_STOP = "/stop"
ENDPOINT_CODE = ["/code", "/code/", "/code/<path:path>", "/code<ending:re:\\.zip>"]
REDIRECT_TO = "http://localhost:8080"
APPLICATION = "schul_cloud_search_tests"
HERE = os.path.dirname(__file__)
LISTING_TEMPLATE = SimpleTemplate(
"""
% import os
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
  </head>
  <body>
    <h1>{{ APPLICATION }}</h1>
    <p>
      You can view the updated source code on
      <a href="https://github.com/schul-cloud/schul_cloud_search_tests">GitHub</a>, create issues and get in touch with the developers.
      You can find the currently running version below.
      You can <a href="/code.zip">download the currently running version of the service</a>.
    </p>
    <ul>
    % for dirpath, dirnames, filenames in os.walk(here):
      % assert dirpath.startswith(here)
      % dirpath = dirpath[len(here):]
      % dirpath = dirpath.replace("\\\\", "/")
      % for filename in filenames:
        % if filename.endswith(".pyc"):
          % continue
        % end
        % file = dirpath + "/" + filename
        <li>
          <a href="/code{{ file }}">{{ file }}</a>
        </li>      
      % end
    % end
    </ul>
    <p>
    </p>
  </body>
</html>
""")


def pytest_errors(status, errors, server_url, answer=None, secret=""):
    """Return the formatted pytest errors, jsonapi compatible."""
    response.status = status
    error_text = ("Bad Request" if status == 400 else "Conflict")
    code = "http://" + request.headers["host"] + "/code"
    result = {
      "errors":[
        {
          "status": str(status),
          "title": error_text,
          "detail": "The request or response contained some errors.",
          "meta": {
            "url": server_url,
            "response": answer
          }
        }
      ] + errors,
      "jsonapi": {
        "version": "1.0",
        "meta": {
          "name": "schul_cloud/schul_cloud_search_tests", 
          "source": code,
          "description":
            "These are the tests for the search engines.",
          "secret": secret
        }
      }
    }
    return result


def check_response(target_url, secret=""):
    """Test the request and the response to the search engine."""
    print("query string:", request.query_string)
    server_url = target_url + "?" + request.query_string
    client_errors = run_request_tests(server_url)
    return_error = (400 if client_errors else 409)
    answer = requests.get(server_url, headers=request.headers)
    server_errors = run_response_tests(server_url, answer)
    try:
        result = answer.json()
    except (JSONDecodeError):
        result = None
    all_errors = []
    if client_errors and answer.status_code != 400:
        all_errors += client_errors
    if server_errors:
        all_errors += server_errors
    if all_errors:
        return pytest_errors(return_error, all_errors, server_url, result,
                             secret=secret)
    assert result is not None, "The tests take care that there is a result."
    response.status = answer.status_code
    response.headers.update(answer.headers)
    return json.dumps(result).encode("UTF-8")


def get_code(path=None, ending=None):
    """Return a directory listing or a static file."""
    if ending is not None:
        assert ending == ".zip"
        redirect("/code/{}.zip".format(APPLICATION))
    if path == APPLICATION + ".zip":
        # Download the source of this application."""
        # from http://stackoverflow.com/questions/458436/adding-folders-to-a-zip-file-using-python#6511788
        zip_path = (shutil.make_archive("/tmp/" + APPLICATION, "zip", HERE))
        return static_file(zip_path, root="/")
    if path is None:
        path = "/"
    if path.endswith("/"):
        # return a listing
        allowed = ["/", "tests/", "search_tests/"]
        assert path in allowed, "Path \"{}\" should be in {}".format(path, allowed)
        return LISTING_TEMPLATE.render(here=HERE, APPLICATION=APPLICATION)
    return static_file(path, root=HERE)


def get_app(endpoint="/", target_url="http://localhost:8080", secret=""):
    """Return a bottle app that tests the request and the response."""
    app = Bottle()
    app.get(endpoint, callback=lambda: check_response(target_url, secret=secret))
    for code_endpoint in ENDPOINT_CODE:
        app.get(code_endpoint, callback=get_code)
    return app


def run(app, host="0.0.0.0", port=8081):
    """Run a stoppable bottle app."""
    server = StoppableWSGIRefServerAdapter(host=host, port=port)
    app.get(ENDPOINT_STOP, callback=lambda: server.shutdown(blocking=False))
    app.run(debug=True, server=server)


def main(host="0.0.0.0", port=8081, endpoint="/", target_url="http://localhost:8080"):
    """Start the server."""
    app = get_app(endpoint=endpoint, target_url=target_url)
    run(app, host=host, port=port)


__all__ = ["main", "run", "get_app"]

if __name__ == "__main__":
    kw = {"host": "0.0.0.0"}
    kw["target_url"] = (sys.argv[1] if len(sys.argv) >= 2 else "http://localhost:8080/v1/search")
    kw["port"] = (int(sys.argv[2]) if len(sys.argv) >= 3 else 8081) 
    kw["endpoint"] = (sys.argv[3] if len(sys.argv) >= 4 else "/v1/search")
    print("Forwarding traffic from http://{host}:{port}{endpoint} to {target_url}".format(**kw))
    main(**kw)



"""These tests test the links which are available.

"""
from pytest import mark
import copy
from schul_cloud_search_tests.tests.assertions import (
    assertServerReplyIsWrong)
from pprint import pprint
import pytest


def test_correct_self_link_is_passed_through(linked_search):
    """Test that the linked search passes through the search engine."""
    for search in linked_search:
        result = search.request().json()
        response = search.response
        print("result:  ", result)
        print("response:", response)
        assert result == response


def test_the_first_response_must_have_the_offest_0(linked_search):
    """The first request is done without specifying an offset.
    The offset 0 is implied."""
    linked_search[0].response["links"]["self"]["meta"]["offset"] = 1
    assertServerReplyIsWrong(linked_search[0].request())


def test_the_limit_may_be_reduced_by_the_server(
        second_search, search_engine, limit):
    """The server can reduce the limit."""
    response = second_search.response
    second_search.parameters["page[limit]"] = str(limit + 1)
    result = search_engine.host(
         second_search.response, second_search.parameters
         ).request().json()
    assert result == response


def test_the_limit_can_not_be_increased_by_the_server(second_search):
    """The server may not increase the limit."""
    second_search.response["links"]["self"]["meta"]["limit"] += 1
    assertServerReplyIsWrong(second_search.request())
    

def test_count_does_not_match(first_search):
    """count must be equal t the number of objects."""
    first_search.response["links"]["self"]["meta"]["count"] += 1
    assertServerReplyIsWrong(first_search.request())


def test_count_can_not_be_greater_than_the_limit(first_search):
    """We reduce the limit so it is smaller than the count. This must be wrong."""
    first_search.response["links"]["self"]["meta"]["limit"] = \
        first_search.response["links"]["self"]["meta"]["count"] - 1
    assertServerReplyIsWrong(first_search.request())
    

def test_count_must_equal_limit_for_links_in_between(not_last_search):
    """If a search is not the last, the count must be equal to the limit.
    
    Test that
    - last last link is self
    """
    response = not_last_search.response
    response["data"].pop()
    links = response["links"]
    meta = links["self"]["meta"]
    meta["count"] -= 1
    links["next"] = None
    assert links["self"]["href"] != links["last"], "precondition for the test"
    assertServerReplyIsWrong(not_last_search.request())


def test_next_link_of_last_request_must_be_null(last_search):
    """If the next link is set in the last request, it is an error."""
    last_search.response["links"]["next"] = \
        last_search.response["links"]["self"]["href"]
    print('last_search.response["links"]:', last_search.response["links"])
    assertServerReplyIsWrong(last_search.request())


@mark.parametrize("link_name", ["first", "last", "prev", "next"])
def test_no_resources_imply_no_links_to_other_searches(first_search, link_name):
    if len(first_search.resources) != 0:
        pytest.skip("Need something with no resources")
    first_search.response["links"][link_name] = \
        first_search.response["links"]["self"]["href"]
    print(first_search.response)
    assertServerReplyIsWrong(first_search.request())


@mark.parametrize("link_name", ["first", "last"])
def test_search_with_offset_too_high_can_have_links(high_offset_search, first_search, link_name):
    """If a search requests an offset too high,
    the first and the last link may be set.
    """
    high_offset_search.response["links"][link_name] = \
        first_search.response["links"]["self"]["href"]
    result = high_offset_search.request().json()
    response = high_offset_search.response
    assert result == response, "search results with too high offset are passed through"


@mark.parametrize("link_name", ["prev", "next"])
def test_search_with_offset_too_high_can_not_have_links(high_offset_search, first_search, link_name):
    """If a search requests an offset too high,
    the prev and the next link can not be set.
    """
    high_offset_search.response["links"][link_name] = \
        first_search.response["links"]["self"]["href"]
    assertServerReplyIsWrong(high_offset_search.request())


def test_last_link_implies_next_link_for_all_but_last_request(
        not_last_search):
    """A last link is given: we test that the next link must exist."""
    assert not_last_search.response["links"]["last"], "precondition for the test"
    not_last_search.response["links"]["next"] = None
    assertServerReplyIsWrong(not_last_search.request())


def test_first_implies_prev_link_for_all_but_first_request(second_search):
    """If the first link is given, also a prev link must be given.
    
    This is only valid, if we did not request an offset which is out of range.
    """
    assert second_search.response["links"]["first"], "precondition for the test"
    second_search.response["links"]["prev"] = None
    assertServerReplyIsWrong(second_search.request())
    

def test_next_link_does_not_skip_objects(first_search, second_search):
    """The next link must follow directly."""
    first_search.response["links"]["next"] = second_search.response["links"]["next"]
    assertServerReplyIsWrong(first_search.request())


def test_prev_link_does_not_skip_objects(second_search, third_search):
    """The prev link must precede directly."""
    third_search.response["links"]["prev"] = second_search.response["links"]["prev"]
    assertServerReplyIsWrong(third_search.request())


def test_href_object_links_are_ok_for_the_search(linked_search):
    """If the links are given, the search must be ok with the href."""
    for search in linked_search:
        response = search.response
        for link_name in ["next", "prev", "first", "last"]:
            if response["links"][link_name]:
                response["links"][link_name] = {
                    "href":response["links"][link_name],
                    "meta": {}}
        result = search.request().json()
        print("result:  ", result)
        print("response:", response)
        assert result == response


@mark.parametrize("link_name", ["self", "prev", "first", "last", "next"])
def test_relative_links_are_not_allowed(second_search, third_search, link_name):
    """Test that the links must start with http or https."""
    existing_link = second_search.response["links"][link_name]
    START = "http://"
    if isinstance(existing_link, str):
        assert existing_link.startswith(START)
        new_link = existing_link[len(START):]
        second_search.response["links"][link_name] = new_link
    if isinstance(existing_link, dict):
        assert existing_link['href'].startswith(START)
        new_link = existing_link['href'][len(START):]
        existing_link['href'] = new_link
    assertServerReplyIsWrong(second_search.request())


def test_requests_beyond_the_last_must_not_have_objects(
        first_search, second_search):
    """A request beyond the last link, there shall be no objects to return.
    
    Make the second search look like the last.
    """
    second_search.response["links"]["next"] = None
    second_search.response["links"]["prev"] = None
    second_search.response["links"]["last"] = first_search.response["links"]["first"]
    assertServerReplyIsWrong(second_search.request())



"""
Each test function here covers a part of the specification.
All tests in here should be marked by 

    @mark.request

They are run on the clients request.
If they fail, the client gets feedback and the server does not get the request.

About the test functions:
Please provide
- A short description of what is tested
- Links to the specification, where this is specified.

"""
from pytest import mark
import pytest
import re


@mark.request
def test_request_has_query(params):
    """Test that the query includes a "Q".
    
    This is specified here:
    - https://github.com/schul-cloud/resources-api-v1/#search-api
    """
    assert params.get("Q") is not None, (
        "Please pass a query string to with Q=... to the request.")


@mark.request
@mark.parametrize("parameter", ["page[offset]", "page[limit]"])
def test_integer_parameters(params, parameter):
    """Test that page[offset] and page[limit] are positive integers.
    
    This is specified here:
    - https://github.com/schul-cloud/resources-api-v1/#search-api
    """
    value = params.get(parameter)
    if value is None:
        pytest.skip("parameter {} absent".format(parameter))
    assert value.isdigit(), (
        "The parameter {} must be a positive integer, not {}."
        .format(parameter, value))
    

JSONAPI_ATTRIBUTE = re.compile("^(sort|filter\\[.+\\]|page\\[(offset|limit)\\])$")


@mark.request
def test_all_parameter_names_are_jsonapi_compatible(params):
    """If search engines add new parameters, they MUST be jsonapi compatible.
    
    See jsonapi
    - http://jsonapi.org/format/#query-parameters
    
    This includes:
    - parameter names
    - filter[ATTRIBUTE.XX.YY....]
    """
    for parameter in params:
        assert parameter, "The parameter must be named \"{}\".".format(parameter)
        for character in (parameter[0], parameter[-1]):
            assert character not in ["-", "_", " "], "Additionally, the following characters are allowed in member names, except as the first or last character: U+002D HYPHEN-MINUS, “-“, U+005F LOW LINE, “_”, and U+0020 SPACE, “ “ (not recommended, not URL safe)"
        has_non_az_character = False
        is_jsonapi_parameter = bool(JSONAPI_ATTRIBUTE.match(parameter))
        for character in parameter:
            assert is_jsonapi_parameter or character not in '+,.[]!"#$%&\'()*/:;<=>?@\\^`{|}~\x7f\x00\x1f', "The following characters MUST NOT be used in member names: " + '\'+\', \',\', \'.\', \'[\', \']\', \'!\', \'"\', \'#\', \'$\', \'%\', \'&\', "\'", \'(\', \')\', \'*\', \'/\', \':\', \';\', \'<\', \'=\', \'>\', \'?\', \'@\', \'\\\\\', \'^\', \'`\', \'{\', \'|\', \'}\', \'~\', \'\\x7f\', \'\\x00\', \'\\x1f\''
            has_non_az_character = has_non_az_character or character < "a"
            has_non_az_character = has_non_az_character or "z" < character
        assert is_jsonapi_parameter or has_non_az_character, "Implementation specific query parameters MUST adhere to the same constraints as member names with the additional requirement that they MUST contain at least one non a-z character (U+0061 to U+007A). Parameter \"{}\"".format(parameter)



from schul_cloud_resources_api_v1.schema import get_schemas
from pytest import mark, fixture
from schul_cloud_search_tests.tests.assertions import (
    assertClientRequestIsInvalid, Q)
from pprint import pprint


@mark.parametrize("param", ["page[offset]", "page[limit]"])
@mark.parametrize("value", ["", "asd", "-1"])
def test_parameter_must_be_positive_integer(request400, assertIs400, param, value):
    """
    According to
    - https://github.com/schul-cloud/resources-api-v1#search-api
    - http://jsonapi.org/format/#fetching-pagination
    page[offset] and page[limit] must be positive integers.
    
    This is an invalid request, according to
    - https://github.com/schul-cloud/schul_cloud_search_tests#specification
    the return code must be ERROR_CLIENT_REQUEST.
    """
    result = request400({param:value, Q:"test"})
    assertIs400(result)
    
    
def test_q_is_a_required_query(search_engine):
    """
    According to the Search API https://github.com/schul-cloud/resources-api-v1#search-api
    q is required
    """
    result = search_engine.request(params={})
    data = result.json()
    assertClientRequestIsInvalid(result)
    errors = [error for error in data["errors"]
             if error["meta"].get("test_function") == "test_request_has_query"]
    assert errors, "There should be such an error function"
    meta = errors[0]["meta"]
    url_start = search_engine.proxy_url + "code/search_tests/test_request.py#"
    assert meta["source"].startswith(url_start)
    from schul_cloud_search_tests.search_tests.test_request import test_request_has_query
    assert meta["source-line"] == test_request_has_query.__code__.co_firstlineno


MALFORMED_PARAMETERS = [
        ("asd", False),
        ("", False),
        ("A", True),
        ("As", True),
        ("aS", True),
        ("aSa", True),
        ("a-a", True),
        ("start- _abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ", True),
        ("asd-a", True),
        ("asd a", True),
        ("asd_a", True),
        ("asd-", False),
        ("asd ", False),
        ("asd_", False),
        ("page[offset]", True),
        ("page[limit]", True),
        ("page[asdas]", False),
        ("sort", True),
        ("filter[asdsa]", True),
        ("filter[]", False),
        ("filter[a.b]", True),
        ("filter[", False),
        ("родина", True),
        ("родина", True),
    ]


@mark.current
@mark.parametrize("parameter,parameter_is_correct", MALFORMED_PARAMETERS)
def test_all_parameter_names_are_jsonapi_compatible(
        search_engine, parameter, parameter_is_correct):
    """make sure the valid and invalid parameters are fed back."""
    response = search_engine.request({parameter:"0", Q:"test"})
    json = response.json()
    pprint(json)
    if parameter_is_correct:
        assert json == search_engine.last_response
    else:
        assertClientRequestIsInvalid(response)




"""
For AGPL compatibility, we want to test the ability to view the code.
This is especially useful if we want to debug the application.
Suppose a test runs locally but not in the container.
There, we need to examine the versions of the tests.

The endpoints are
/source/...
  with listings and files
/source.zip -> /schul_cloud_search_tests.zip
  a zip file of the source code
"""
import requests
from pytest import mark
import os

HERE = os.path.dirname(__file__)


@mark.parametrize("directory,content", [
        ("/", ["proxy.py"]),
        ("/tests/", ["__init__.py", "conftest.py", "test_code.py", "README.rst"]),
        ("/search_tests/", ["__init__.py", "conftest.py", "test_links.py", "README.rst"]),
    ])
@mark.parametrize("appendix", ["", "/"])
def test_directories_are_listed(code_url, appendix, directory, content):
    """Make sure the files listed to view them."""
    text = requests.get(code_url + appendix).text
    print(text)
    for entry in content:
        included_text = "<a href=\"/code{}{}\"".format(directory, entry)
        assert included_text in text


@mark.parametrize("file,url", [
        ("test_code.py", "/tests/test_code.py"),
        ("../proxy.py", "/proxy.py")
    ])
def test_can_download_files(code_url, file, url):
    """Download the zip file of all files."""
    with open(os.path.join(HERE, file)) as f:
        expected_text = f.read()
    response = requests.get(code_url + url)
    assert response.text == expected_text


def test_can_download_joined_zip_file(code_url):
    """The request to /code.zip redirects to the package name zip file."""
    response = requests.get(code_url + ".zip")
    assert response.url.endswith("schul_cloud_search_tests.zip")


def test_listing_links_to_repository(code_url):
    """The link to the repository is included in the listing."""
    response = requests.get(code_url)
    assert "<a href=\"https://github.com/schul-cloud/schul_cloud_search_tests\"" in response.text


def test_source_code_download_link_is_in_listing(code_url):
    """Make sure users can download the current code."""
    response = requests.get(code_url)
    assert "<a href=\"/code.zip\"" in response.text


def test_source_link_of_search_engine_refers_to_source_endpoint(
        search_engine, code_url):
    """Test that the source link of the response of the search engine
    refers to the code enpoint /code of the same search engine."""
    response = search_engine.request({})
    source = response.json()["jsonapi"]["meta"]["source"]
    assert source == code_url



"""
This file contains the test that validates the result.


"""
from pytest import mark


def get_href(link):
    """Return the reference of the link."""
    if link is None:
        return link
    if isinstance(link, dict):
        return link.get("href")
    return link


@mark.current
def test_self_link_works_multiple_times(query, validateRequest):
    """The self link can be requested again.
    
    At least the self link should be the same.
    However here, we assume the same response is given.
    """
    first_result = validateRequest(query).json()
    href = first_result["links"]["self"]["href"]
    second_result = validateRequest(href).json()
    assert first_result == second_result


def test_linked_resources(query, validateRequest, max_depth):
    """Test that the linked resources all link to eachother."""
    results = [validateRequest(query)]
    for i in range(max_depth):
        last_result = results[-1]
        next_link = get_href(last_result.json()["links"]["next"])
        if next_link is not None:
            results.append(validateRequest(next_link))
    for self, next in zip(results, results[1:]):
        next_self = get_href(next.json()["links"]["self"])
        self_self = get_href(self.json()["links"]["self"])
        next_prev = get_href(next.json()["links"]["prev"])
        self_next = get_href(self.json()["links"]["next"])
        assert next_self == self_next, "The self link points to the next link."
        assert self_self == next_prev, "The next link points to the self link."
    last_links = set(get_href(link.json()["links"]["last"]) for link in results)
    first_links = set(get_href(link.json()["links"]["first"]) for link in results)
    assert len(first_links) == 1, "All results should have the same first link."
    assert len(last_links) == 1, "All results should have the same last link."


@mark.parametrize("sort", ["title", "title,url", "-url,title"])
def test_sorting(validateRequest, sort):
    """Test the sort parameter as described by the jsonapi.
    
    http://jsonapi.org/format/#fetching-sorting
    """
    validateRequest({"Q": "test", "sort": sort})


def test_filtering(validateRequest):
    """Test the filtering as described by the json api and search api.
    
    See
    - https://github.com/schul-cloud/resources-api-v1/#search-api
    - http://jsonapi.org/format/#fetching-filtering
    """
    validateRequest({"Q": "test", "filter[title]": "test"})
    



"""
This file contains the assertions for the search tests.
"""
from schul_cloud_search_tests.tests.assertions import (
    assertClientRequestIsInvalid)


def assertSearchEngineExitedWith400(response):
    """Make sure that the response is from the client and not the tests.
    
    The error should be 400.
    """





from schul_cloud_resources_api_v1.schema import get_schemas
from pytest import mark
from schul_cloud_search_tests.tests.assertions import (
    assertIsError, ERROR_CLIENT_REQUEST, Q, ERROR_SERVER_RESPONSE, ERROR)
from pprint import pprint
import copy


@mark.parametrize("response", get_schemas()["search-response"].get_valid_examples())
def test_valid_query_is_returned(search_engine, response):
    """Valid results are passed through the search engine"""
    offset = response["links"]["self"]["meta"]["offset"]
    query = {Q:"test2", "page[offset]":str(offset)}
    result = search_engine.host(response, query).request()
    data = result.json()
    pprint(data)
    assert data == response


@mark.parametrize("response", get_schemas()["search-response"].get_invalid_examples())
def test_invalid_response_is_detected(search_engine, response):
    """Detect when the search engine returns a response
    which does not fit the schema
    """
    result = search_engine.host(response).request()
    assertIsError(result, ERROR_SERVER_RESPONSE)


@mark.parametrize("invalid_response", ["{", "", b"asd", "{}", "[]"])
def test_invalid_return_values_are_handled(search_engine, invalid_response):
    """Test that a invalid responses still create a useful output.
    
    - invalid json
    - valid json but too small
    - bytes
    """
    result = search_engine.host(invalid_response).request()
    assertIsError(result, ERROR_SERVER_RESPONSE)


@mark.parametrize("content_type", [
    "application/json", "ajhkjf", "application/vnd.api+json1"])
def test_invalid_return_header_type_is_handled(search_engine, content_type):
    """The header must be correct.
    
    Wrong headers:
    - with parameters
    - application/json
    """
    result = search_engine.host(headers={"Content-Type":content_type}).request()
    assertIsError(result, ERROR_SERVER_RESPONSE)

@mark.parametrize("code,name", [
        (400, 'Bad Request'),
        (401, 'Unauthorized'), 
        (402, 'Payment Required'),
        (405, 'Method Not Allowed'),
        (409, 'Conflict'),
        (500, 'Internal Server Error'),
        (501, 'Not Implemented'),
        (503, 'Service Unavailable'),
        (504, 'Gateway Timeout'),
    ])
def test_400_and_500_status_codes_pass_through(
        search_engine, code, name):
    """Test that the error has the right format and is passed through directly."""
    error = copy.deepcopy(ERROR)
    error["errors"][0]["status"] = str(code)
    error["errors"][0]["title"] = name
    assertIsError(error, code)
    response = search_engine.host(error, status_code=code).request()
    assert response.status_code == code
    assert error == response.json()


@mark.parametrize("code", [200, 201, 301, 466, 588])
def test_status_code_and_error_code_must_match(search_engine, code):
    """The serach engine detects of the returned error
    and the status code do not match.
    """
    result = search_engine.host(ERROR, status_code=code).request()
    assertIsError(result, ERROR_SERVER_RESPONSE)

